@@ -29,6 +29,7 @@ the `session` module provides the main user-facing abstractions.
 __all__ = [
     "Bolt",
     "BoltPool",
+    "ConnectionErrorHandler",
     "Neo4jPool",
     "check_supported_server_product",
 ]
@@ -38,6 +39,7 @@ from collections import (
     defaultdict,
     deque,
 )
+import logging
 from logging import getLogger
 from random import choice
 import selectors
@@ -57,7 +59,6 @@ from ssl import (
 )
 from threading import (
     Condition,
-    Lock,
     RLock,
 )
 from time import perf_counter
@@ -66,7 +67,6 @@ from neo4j._exceptions import (
     BoltError,
     BoltHandshakeError,
     BoltProtocolError,
-    BoltRoutingError,
     BoltSecurityError,
 )
 from neo4j.addressing import Address
@@ -95,6 +95,7 @@ from neo4j.exceptions import (
 )
 from neo4j.io._common import (
     CommitResponse,
+    ConnectionErrorHandler,
     Inbox,
     InitResponse,
     Outbox,
@@ -126,7 +127,7 @@ class Bolt(abc.ABC):
     PROTOCOL_VERSION = None
 
     # flag if connection needs RESET to go back to READY state
-    _is_reset = True
+    is_reset = False
 
     # The socket
     in_use = False
@@ -140,6 +141,11 @@ class Bolt(abc.ABC):
     #: The pool of which this connection is a member
     pool = None
 
+    # Store the id of the most recent ran query to be able to reduce sent bits by
+    # using the default (-1) to refer to the most recent query when pulling
+    # results for it.
+    most_recent_qid = None
+
     def __init__(self, unresolved_address, sock, max_connection_lifetime, *, auth=None, user_agent=None, routing_context=None):
         self.unresolved_address = unresolved_address
         self.socket = sock
@@ -218,7 +224,7 @@ class Bolt(abc.ABC):
 
         # Carry out Bolt subclass imports locally to avoid circular dependency issues.
         from neo4j.io._bolt3 import Bolt3
-        from neo4j.io._bolt4 import Bolt4x0, Bolt4x1, Bolt4x2, Bolt4x3
+        from neo4j.io._bolt4 import Bolt4x0, Bolt4x1, Bolt4x2, Bolt4x3, Bolt4x4
 
         handlers = {
             Bolt3.PROTOCOL_VERSION: Bolt3,
@@ -226,6 +232,7 @@ class Bolt(abc.ABC):
             Bolt4x1.PROTOCOL_VERSION: Bolt4x1,
             Bolt4x2.PROTOCOL_VERSION: Bolt4x2,
             Bolt4x3.PROTOCOL_VERSION: Bolt4x3,
+            Bolt4x4.PROTOCOL_VERSION: Bolt4x4,
         }
 
         if protocol_version is None:
@@ -245,25 +252,24 @@ class Bolt(abc.ABC):
         preference. The number of protocol versions (or ranges)
         returned is limited to four.
         """
-        ranges_supported = versions[0] >= Version(4, 3)
-        without_changes = {Version(4, 2)}
-        if versions and ranges_supported:
-            start, end = 0, 0
-            first_major = versions[start][0]
-            minors = []
-            for version in versions:
-                if version[0] == first_major:
-                    minors.append(version[1])
-                else:
-                    break
-            new_versions = [
-                Version(first_major, minors),
-                *(v for v in versions[1:] if v not in without_changes)
-            ]
-        else:
-            new_versions = [v for v in versions if v not in without_changes]
-        new_versions = [*new_versions[:(limit - 1)], versions[-1]]
-        return new_versions
+        # In fact, 4.3 is the fist version to support ranges. However, the range
+        # support got backported to 4.2. But even if the server is too old to
+        # have the backport, negotiating BOLT 4.1 is no problem as it's
+        # equivalent to 4.2
+        first_with_range_support = Version(4, 2)
+        result = []
+        for version in versions:
+            if (result
+                    and version >= first_with_range_support
+                    and result[-1][0] == version[0]
+                    and result[-1][1][1] == version[1] + 1):
+                # can use range to encompass this version
+                result[-1][1][1] = version[1]
+                continue
+            result.append(Version(version[0], [version[1], version[1]]))
+            if len(result) == 4:
+                break
+        return result
 
     @classmethod
     def get_handshake(cls):
@@ -317,26 +323,26 @@ class Bolt(abc.ABC):
             keep_alive=pool_config.keep_alive,
         )
 
+        # Carry out Bolt subclass imports locally to avoid circular dependency
+        # issues.
         if pool_config.protocol_version == (3, 0):
-            # Carry out Bolt subclass imports locally to avoid circular dependency issues.
             from neo4j.io._bolt3 import Bolt3
-            connection = Bolt3(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
+            bolt_cls = Bolt3
         elif pool_config.protocol_version == (4, 0):
-            # Carry out Bolt subclass imports locally to avoid circular dependency issues.
             from neo4j.io._bolt4 import Bolt4x0
-            connection = Bolt4x0(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
+            bolt_cls = Bolt4x0
         elif pool_config.protocol_version == (4, 1):
-            # Carry out Bolt subclass imports locally to avoid circular dependency issues.
             from neo4j.io._bolt4 import Bolt4x1
-            connection = Bolt4x1(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
+            bolt_cls = Bolt4x1
         elif pool_config.protocol_version == (4, 2):
-            # Carry out Bolt subclass imports locally to avoid circular dependency issues.
             from neo4j.io._bolt4 import Bolt4x2
-            connection = Bolt4x2(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
+            bolt_cls = Bolt4x2
         elif pool_config.protocol_version == (4, 3):
-            # Carry out Bolt subclass imports locally to avoid circular dependency issues.
             from neo4j.io._bolt4 import Bolt4x3
-            connection = Bolt4x3(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
+            bolt_cls = Bolt4x3
+        elif pool_config.protocol_version == (4, 4):
+            from neo4j.io._bolt4 import Bolt4x4
+            bolt_cls = Bolt4x4
         else:
             log.debug("[#%04X]  S: <CLOSE>", s.getsockname()[1])
             _close_socket(s)
@@ -344,6 +350,11 @@ class Bolt(abc.ABC):
             supported_versions = Bolt.protocol_handlers().keys()
             raise BoltHandshakeError("The Neo4J server does not support communication with this driver. This driver have support for Bolt Protocols {}".format(supported_versions), address=address, request_data=handshake, response_data=data)
 
+        connection = bolt_cls(
+            address, s, pool_config.max_connection_lifetime, auth=auth,
+            user_agent=pool_config.user_agent, routing_context=routing_context
+        )
+
         try:
             connection.hello()
         except Exception:
@@ -381,7 +392,7 @@ class Bolt(abc.ABC):
             pass
 
     @abc.abstractmethod
-    def route(self, database=None, bookmarks=None):
+    def route(self, database=None, imp_user=None, bookmarks=None):
         """ Fetch a routing table from the server for the given
         `database`. For Bolt 4.3 and above, this appends a ROUTE
         message; for earlier versions, a procedure call is made via
@@ -389,6 +400,7 @@ class Bolt(abc.ABC):
         sent to the network, and a response is fetched.
 
         :param database: database for which to fetch a routing table
+        :param imp_user: the user to impersonate
         :param bookmarks: iterable of bookmark values after which this
                           transaction should begin
         :return: dictionary of raw routing data
@@ -396,8 +408,8 @@ class Bolt(abc.ABC):
         pass
 
     @abc.abstractmethod
-    def run(self, query, parameters=None, mode=None, bookmarks=None, metadata=None,
-            timeout=None, db=None, **handlers):
+    def run(self, query, parameters=None, mode=None, bookmarks=None,
+            metadata=None, timeout=None, db=None, imp_user=None, **handlers):
         """ Appends a RUN message to the output queue.
 
         :param query: Cypher query string
@@ -407,6 +419,7 @@ class Bolt(abc.ABC):
         :param metadata: custom metadata dictionary to attach to the transaction
         :param timeout: timeout for transaction execution (seconds)
         :param db: name of the database against which to begin the transaction
+        :param imp_user: the user to impersonate
         :param handlers: handler functions passed into the returned Response object
         :return: Response object
         """
@@ -435,7 +448,8 @@ class Bolt(abc.ABC):
         pass
 
     @abc.abstractmethod
-    def begin(self, mode=None, bookmarks=None, metadata=None, timeout=None, db=None, **handlers):
+    def begin(self, mode=None, bookmarks=None, metadata=None, timeout=None,
+              db=None, imp_user=None, **handlers):
         """ Appends a BEGIN message to the output queue.
 
         :param mode: access mode for routing - "READ" or "WRITE" (default)
@@ -443,6 +457,7 @@ class Bolt(abc.ABC):
         :param metadata: custom metadata dictionary to attach to the transaction
         :param timeout: timeout for transaction execution (seconds)
         :param db: name of the database against which to begin the transaction
+        :param imp_user: the user to impersonate
         :param handlers: handler functions passed into the returned Response object
         :return: Response object
         """
@@ -458,10 +473,6 @@ class Bolt(abc.ABC):
         """ Appends a ROLLBACK message to the output queue."""
         pass
 
-    @property
-    def is_reset(self):
-        return self._is_reset
-
     @abc.abstractmethod
     def reset(self):
         """ Appends a RESET message to the outgoing queue, sends it and consumes
@@ -476,9 +487,9 @@ class Bolt(abc.ABC):
         :param fields: the fields of the message as a tuple
         :param response: a response object to handle callbacks
         """
-        self.packer.pack_struct(signature, fields)
-        self.outbox.chunk()
-        self.outbox.chunk()
+        with self.outbox.tmp_buffer():
+            self.packer.pack_struct(signature, fields)
+        self.outbox.wrap_message()
         self.responses.append(response)
 
     def _send_all(self):
@@ -643,11 +654,19 @@ class IOPool:
                 # try to find a free connection in pool
                 for connection in list(self.connections.get(address, [])):
                     if (connection.closed() or connection.defunct()
-                            or connection.stale()):
+                            or (connection.stale() and not connection.in_use)):
                         # `close` is a noop on already closed connections.
                         # This is to make sure that the connection is gracefully
                         # closed, e.g. if it's just marked as `stale` but still
                         # alive.
+                        if log.isEnabledFor(logging.DEBUG):
+                            log.debug(
+                                "[#%04X]  C: <POOL> removing old connection "
+                                "(closed=%s, defunct=%s, stale=%s, in_use=%s)",
+                                connection.local_port,
+                                connection.closed(), connection.defunct(),
+                                connection.stale(), connection.in_use
+                            )
                         connection.close()
                         try:
                             self.connections.get(address, []).remove(connection)
@@ -715,9 +734,9 @@ class IOPool:
         """
         with self.lock:
             for connection in connections:
-                if not (connection.is_reset
-                        or connection.defunct()
-                        or connection.closed()):
+                if not (connection.defunct()
+                        or connection.closed()
+                        or connection.is_reset):
                     try:
                         connection.reset()
                     except (Neo4jError, DriverError, BoltError) as e:
@@ -818,8 +837,7 @@ class BoltPool(IOPool):
     def __repr__(self):
         return "<{} address={!r}>".format(self.__class__.__name__, self.address)
 
-    def acquire(self, access_mode=None, timeout=None, database=None,
-                bookmarks=None):
+    def acquire(self, access_mode=None, timeout=None, database=None, bookmarks=None):
         # The access_mode and database is not needed for a direct connection, its just there for consistency.
         return self._acquire(self.address, timeout)
 
@@ -860,14 +878,14 @@ class Neo4jPool(IOPool):
         :param opener:
         :param pool_config:
         :param workspace_config:
-        :param addresses:
+        :param address:
         """
         super(Neo4jPool, self).__init__(opener, pool_config, workspace_config)
         # Each database have a routing table, the default database is a special case.
         log.debug("[#0000]  C: <NEO4J POOL> routing address %r", address)
         self.address = address
         self.routing_tables = {workspace_config.database: RoutingTable(database=workspace_config.database, routers=[address])}
-        self.refresh_lock = Lock()
+        self.refresh_lock = RLock()
 
     def __repr__(self):
         """ The representation shows the initial routing addresses.
@@ -900,15 +918,24 @@ class Neo4jPool(IOPool):
     def get_routing_table_for_default_database(self):
         return self.routing_tables[self.workspace_config.database]
 
-    def create_routing_table(self, database):
-        if database not in self.routing_tables:
-            self.routing_tables[database] = RoutingTable(database=database, routers=self.get_default_database_initial_router_addresses())
+    def get_or_create_routing_table(self, database):
+        with self.refresh_lock:
+            if database not in self.routing_tables:
+                self.routing_tables[database] = RoutingTable(
+                    database=database,
+                    routers=self.get_default_database_initial_router_addresses()
+                )
+            return self.routing_tables[database]
 
-    def fetch_routing_info(self, address, database, bookmarks, timeout):
+    def fetch_routing_info(self, address, database, imp_user, bookmarks,
+                           timeout):
         """ Fetch raw routing info from a given router address.
 
         :param address: router address
         :param database: the database name to get routing table for
+        :param imp_user: the user to impersonate while fetching the routing
+                         table
+        :type imp_user: str or None
         :param bookmarks: iterable of bookmark values after which the routing
                           info should be fetched
         :param timeout: connection acquisition timeout in seconds
@@ -918,59 +945,56 @@ class Neo4jPool(IOPool):
         :raise ServiceUnavailable: if the server does not support
             routing, or if routing support is broken or outdated
         """
+        cx = self._acquire(address, timeout)
         try:
-            cx = self._acquire(address, timeout)
-            try:
-                routing_table = cx.route(
-                    database or self.workspace_config.database, bookmarks
-                )
-            finally:
-                self.release(cx)
-        except BoltRoutingError as error:
-            # Connection was successful, but routing support is
-            # broken. This may indicate that the routing procedure
-            # does not exist (for protocol versions prior to 4.3).
-            # This error is converted into ServiceUnavailable,
-            # therefore surfacing to the application as a signal that
-            # routing is broken.
-            log.debug("Routing is broken (%s)", error)
-            raise ServiceUnavailable(*error.args)
-        except (ServiceUnavailable, SessionExpired) as error:
-            # The routing table request suffered a connection
-            # failure. This should return a null routing table,
-            # signalling to the caller to retry the request
-            # elsewhere.
-            log.debug("Routing is unavailable (%s)", error)
-            routing_table = None
-        # If the routing table is empty, deactivate the address.
-        if not routing_table:
-            self.deactivate(address)
+            routing_table = cx.route(
+                database or self.workspace_config.database,
+                imp_user or self.workspace_config.impersonated_user,
+                bookmarks
+            )
+        finally:
+            self.release(cx)
         return routing_table
 
-    def fetch_routing_table(self, *, address, timeout, database, bookmarks):
+    def fetch_routing_table(self, *, address, timeout, database, imp_user,
+                            bookmarks):
         """ Fetch a routing table from a given router address.
 
         :param address: router address
         :param timeout: seconds
         :param database: the database name
         :type: str
+        :param imp_user: the user to impersonate while fetching the routing
+                         table
+        :type imp_user: str or None
         :param bookmarks: bookmarks used when fetching routing table
 
         :return: a new RoutingTable instance or None if the given router is
                  currently unable to provide routing information
         """
+        new_routing_info = None
         try:
-            new_routing_info = self.fetch_routing_info(address, database,
-                                                       bookmarks, timeout)
+            new_routing_info = self.fetch_routing_info(
+                address, database, imp_user, bookmarks, timeout
+            )
+        except Neo4jError as e:
+            # checks if the code is an error that is caused by the client. In
+            # this case there is no sense in trying to fetch a RT from another
+            # router. Hence, the driver should fail fast during discovery.
+            if e.is_fatal_during_discovery():
+                raise
         except (ServiceUnavailable, SessionExpired):
-            new_routing_info = None
+            pass
         if not new_routing_info:
             log.debug("Failed to fetch routing info %s", address)
             return None
         else:
             servers = new_routing_info[0]["servers"]
             ttl = new_routing_info[0]["ttl"]
-            new_routing_table = RoutingTable.parse_routing_info(database=database, servers=servers, ttl=ttl)
+            database = new_routing_info[0].get("db", database)
+            new_routing_table = RoutingTable.parse_routing_info(
+                database=database, servers=servers, ttl=ttl
+            )
 
         # Parse routing info and count the number of each type of server
         num_routers = len(new_routing_table.routers)
@@ -993,8 +1017,8 @@ class Neo4jPool(IOPool):
         # At least one of each is fine, so return this table
         return new_routing_table
 
-    def update_routing_table_from(self, *routers, database=None,
-                                  bookmarks=None):
+    def _update_routing_table_from(self, *routers, database=None, imp_user=None,
+                                   bookmarks=None, database_callback=None):
         """ Try to update routing tables with the given routers.
 
         :return: True if the routing table is successfully updated,
@@ -1006,67 +1030,86 @@ class Neo4jPool(IOPool):
                 new_routing_table = self.fetch_routing_table(
                     address=address,
                     timeout=self.pool_config.connection_timeout,
-                    database=database, bookmarks=bookmarks
+                    database=database, imp_user=imp_user, bookmarks=bookmarks
                 )
                 if new_routing_table is not None:
-                    self.routing_tables[database].update(new_routing_table)
+                    new_databse = new_routing_table.database
+                    self.get_or_create_routing_table(new_databse)\
+                        .update(new_routing_table)
                     log.debug(
                         "[#0000]  C: <UPDATE ROUTING TABLE> address=%r (%r)",
-                        address, self.routing_tables[database]
+                        address, self.routing_tables[new_databse]
                     )
+                    if callable(database_callback):
+                        database_callback(new_databse)
                     return True
             self.deactivate(router)
         return False
 
-    def update_routing_table(self, *, database, bookmarks):
+    def update_routing_table(self, *, database, imp_user, bookmarks,
+                             database_callback=None):
         """ Update the routing table from the first router able to provide
         valid routing information.
 
         :param database: The database name
+        :param imp_user: the user to impersonate while fetching the routing
+                         table
+        :type imp_user: str or None
         :param bookmarks: bookmarks used when fetching routing table
+        :param database_callback: A callback function that will be called with
+            the database name as only argument when a new routing table has been
+            acquired. This database name might different from `database` if that
+            was None and the underlying protocol supports reporting back the
+            actual database.
 
         :raise neo4j.exceptions.ServiceUnavailable:
         """
-        # copied because it can be modified
-        existing_routers = set(self.routing_tables[database].routers)
+        with self.refresh_lock:
+            # copied because it can be modified
+            existing_routers = set(
+                self.get_or_create_routing_table(database).routers
+            )
 
-        prefer_initial_routing_address = \
-            self.routing_tables[database].missing_fresh_writer()
+            prefer_initial_routing_address = \
+                self.routing_tables[database].initialized_without_writers
 
-        if prefer_initial_routing_address:
-            # TODO: Test this state
-            if self.update_routing_table_from(
+            if prefer_initial_routing_address:
+                # TODO: Test this state
+                if self._update_routing_table_from(
+                        self.first_initial_routing_address, database=database,
+                        imp_user=imp_user, bookmarks=bookmarks,
+                        database_callback=database_callback
+                ):
+                    # Why is only the first initial routing address used?
+                    return
+            if self._update_routing_table_from(
+                    *(existing_routers - {self.first_initial_routing_address}),
+                    database=database, imp_user=imp_user, bookmarks=bookmarks,
+                    database_callback=database_callback
+            ):
+                return
+
+            if not prefer_initial_routing_address:
+                if self._update_routing_table_from(
                     self.first_initial_routing_address, database=database,
-                    bookmarks=bookmarks
-            ):
-                # Why is only the first initial routing address used?
-                return
-        if self.update_routing_table_from(
-                *(existing_routers - {self.first_initial_routing_address}),
-                database=database, bookmarks=bookmarks
-        ):
-            return
+                    imp_user=imp_user, bookmarks=bookmarks,
+                    database_callback=database_callback
+                ):
+                    # Why is only the first initial routing address used?
+                    return
 
-        if not prefer_initial_routing_address:
-            if self.update_routing_table_from(
-                self.first_initial_routing_address, database=database,
-                bookmarks=bookmarks
-            ):
-                # Why is only the first initial routing address used?
-                return
-
-        # None of the routers have been successful, so just fail
-        log.error("Unable to retrieve routing information")
-        raise ServiceUnavailable("Unable to retrieve routing information")
+            # None of the routers have been successful, so just fail
+            log.error("Unable to retrieve routing information")
+            raise ServiceUnavailable("Unable to retrieve routing information")
 
     def update_connection_pool(self, *, database):
-        servers = self.routing_tables[database].servers()
+        servers = self.get_or_create_routing_table(database).servers()
         for address in list(self.connections):
-            if address not in servers:
+            if address.unresolved not in servers:
                 super(Neo4jPool, self).deactivate(address)
 
-    def ensure_routing_table_is_fresh(self, *, access_mode, database,
-                                      bookmarks):
+    def ensure_routing_table_is_fresh(self, *, access_mode, database, imp_user,
+                                      bookmarks, database_callback=None):
         """ Update the routing table if stale.
 
         This method performs two freshness checks, before and after acquiring
@@ -1080,44 +1123,51 @@ class Neo4jPool(IOPool):
         :return: `True` if an update was required, `False` otherwise.
         """
         from neo4j.api import READ_ACCESS
-        if self.routing_tables[database].is_fresh(readonly=(access_mode == READ_ACCESS)):
-            # Readers are fresh.
-            return False
         with self.refresh_lock:
+            if self.get_or_create_routing_table(database)\
+                    .is_fresh(readonly=(access_mode == READ_ACCESS)):
+                # Readers are fresh.
+                return False
 
-            self.update_routing_table(database=database, bookmarks=bookmarks)
+            self.update_routing_table(
+                database=database, imp_user=imp_user, bookmarks=bookmarks,
+                database_callback=database_callback
+            )
             self.update_connection_pool(database=database)
 
             for database in list(self.routing_tables.keys()):
                 # Remove unused databases in the routing table
                 # Remove the routing table after a timeout = TTL + 30s
                 log.debug("[#0000]  C: <ROUTING AGED> database=%s", database)
-                if self.routing_tables[database].should_be_purged_from_memory() and database != self.workspace_config.database:
+                if (self.routing_tables[database].should_be_purged_from_memory()
+                        and database != self.workspace_config.database):
                     del self.routing_tables[database]
 
             return True
 
-    def _select_address(self, *, access_mode, database, bookmarks):
+    def _select_address(self, *, access_mode, database):
         from neo4j.api import READ_ACCESS
         """ Selects the address with the fewest in-use connections.
         """
-        self.create_routing_table(database)
-        self.ensure_routing_table_is_fresh(
-            access_mode=access_mode, database=database, bookmarks=bookmarks
-        )
-        log.debug("[#0000]  C: <ROUTING TABLE ENSURE FRESH> %r", self.routing_tables)
-        if access_mode == READ_ACCESS:
-            addresses = self.routing_tables[database].readers
-        else:
-            addresses = self.routing_tables[database].writers
-        addresses_by_usage = {}
-        for address in addresses:
-            addresses_by_usage.setdefault(self.in_use_connection_count(address), []).append(address)
+        with self.refresh_lock:
+            if access_mode == READ_ACCESS:
+                addresses = self.routing_tables[database].readers
+            else:
+                addresses = self.routing_tables[database].writers
+            addresses_by_usage = {}
+            for address in addresses:
+                addresses_by_usage.setdefault(
+                    self.in_use_connection_count(address), []
+                ).append(address)
         if not addresses_by_usage:
             if access_mode == READ_ACCESS:
-                raise ReadServiceUnavailable("No read service currently available")
+                raise ReadServiceUnavailable(
+                    "No read service currently available"
+                )
             else:
-                raise WriteServiceUnavailable("No write service currently available")
+                raise WriteServiceUnavailable(
+                    "No write service currently available"
+                )
         return choice(addresses_by_usage[min(addresses_by_usage)])
 
     def acquire(self, access_mode=None, timeout=None, database=None,
@@ -1125,21 +1175,29 @@ class Neo4jPool(IOPool):
         if access_mode not in (WRITE_ACCESS, READ_ACCESS):
             raise ClientError("Non valid 'access_mode'; {}".format(access_mode))
         if not timeout:
-            raise ClientError("'timeout' must be a float larger than 0; {}".format(timeout))
+            raise ClientError("'timeout' must be a float larger than 0; {}"
+                              .format(timeout))
 
         from neo4j.api import check_access_mode
         access_mode = check_access_mode(access_mode)
+        with self.refresh_lock:
+            log.debug("[#0000]  C: <ROUTING TABLE ENSURE FRESH> %r",
+                      self.routing_tables)
+            self.ensure_routing_table_is_fresh(
+                access_mode=access_mode, database=database, imp_user=None,
+                bookmarks=bookmarks
+            )
+
         while True:
             try:
-                # Get an address for a connection that have the fewest in-use connections.
-                address = self._select_address(
-                    access_mode=access_mode, database=database,
-                    bookmarks=bookmarks
-                )
-                log.debug("[#0000]  C: <ACQUIRE ADDRESS> database=%r address=%r", database, address)
+                # Get an address for a connection that have the fewest in-use
+                # connections.
+                address = self._select_address(access_mode=access_mode,
+                                               database=database)
             except (ReadServiceUnavailable, WriteServiceUnavailable) as err:
                 raise SessionExpired("Failed to obtain connection towards '%s' server." % access_mode) from err
             try:
+                log.debug("[#0000]  C: <ACQUIRE ADDRESS> database=%r address=%r", database, address)
                 connection = self._acquire(address, timeout=timeout)  # should always be a resolved address
             except ServiceUnavailable:
                 self.deactivate(address=address)
