@@ -19,9 +19,12 @@
 # limitations under the License.
 
 
+from functools import wraps
+
+
 # Can be automatically overridden in builds
 package = "neo4j"
-version = "4.3.8"
+version = "4.4.3"
 
 
 def get_user_agent():
@@ -34,6 +37,11 @@ def get_user_agent():
     return template.format(*fields)
 
 
+def deprecation_warn(message):
+    from warnings import warn
+    warn(message, category=DeprecationWarning, stacklevel=2)
+
+
 def deprecated(message):
     """ Decorator for deprecating functions and methods.
 
@@ -45,13 +53,10 @@ def deprecated(message):
 
     """
     def f__(f):
+        @wraps(f)
         def f_(*args, **kwargs):
-            from warnings import warn
-            warn(message, category=DeprecationWarning, stacklevel=2)
+            deprecation_warn(message)
             return f(*args, **kwargs)
-        f_.__name__ = f.__name__
-        f_.__doc__ = f.__doc__
-        f_.__dict__.update(f.__dict__)
         return f_
     return f__
 
@@ -73,12 +78,10 @@ def experimental(message):
 
     """
     def f__(f):
+        @wraps(f)
         def f_(*args, **kwargs):
             from warnings import warn
             warn(message, category=ExperimentalWarning, stacklevel=2)
             return f(*args, **kwargs)
-        f_.__name__ = f.__name__
-        f_.__doc__ = f.__doc__
-        f_.__dict__.update(f.__dict__)
         return f_
     return f__
