@@ -77,10 +77,6 @@ class Session(Workspace):
     # The current auto-transaction result, if any.
     _autoResult = None
 
-    # The set of bookmarks after which the next
-    # :class:`.Transaction` should be carried out.
-    _bookmarks = None
-
     # The state this session is in.
     _state_failed = False
 
@@ -95,7 +91,7 @@ class Session(Workspace):
     def __del__(self):
         try:
             self.close()
-        except OSError:
+        except (OSError, ServiceUnavailable, SessionExpired):
             pass
 
     def __enter__(self):
@@ -106,26 +102,10 @@ class Session(Workspace):
             self._state_failed = True
         self.close()
 
-    def _connect(self, access_mode, database):
+    def _connect(self, access_mode):
         if access_mode is None:
             access_mode = self._config.default_access_mode
-        if self._connection:
-            # TODO: Investigate this
-            # log.warning("FIXME: should always disconnect before connect")
-            self._connection.send_all()
-            self._connection.fetch_all()
-            self._disconnect()
-        self._connection = self._pool.acquire(
-            access_mode=access_mode,
-            timeout=self._config.connection_acquisition_timeout,
-            database=database,
-            bookmarks=self._bookmarks
-        )
-
-    def _disconnect(self):
-        if self._connection:
-            self._pool.release(self._connection)
-            self._connection = None
+        super()._connect(access_mode)
 
     def _collect_bookmark(self, bookmark):
         if bookmark:
@@ -137,13 +117,16 @@ class Session(Workspace):
             self._autoResult = None
             self._disconnect()
 
-    def _result_network_error(self):
+    def _result_error(self, _):
         if self._autoResult:
             self._autoResult = None
             self._disconnect()
 
     def close(self):
-        """Close the session. This will release any borrowed resources, such as connections, and will roll back any outstanding transactions.
+        """Close the session.
+
+        This will release any borrowed resources, such as connections, and will
+        roll back any outstanding transactions.
         """
         if self._connection:
             if self._autoResult:
@@ -218,7 +201,7 @@ class Session(Workspace):
             self._autoResult._buffer_all()  # This will buffer upp all records for the previous auto-transaction
 
         if not self._connection:
-            self._connect(self._config.default_access_mode, database=self._config.database)
+            self._connect(self._config.default_access_mode)
         cx = self._connection
         protocol_version = cx.PROTOCOL_VERSION
         server_info = cx.server_info
@@ -227,11 +210,12 @@ class Session(Workspace):
 
         self._autoResult = Result(
             cx, hydrant, self._config.fetch_size, self._result_closed,
-            self._result_network_error
+            self._result_error
         )
         self._autoResult._run(
             query, parameters, self._config.database,
-            self._config.default_access_mode, self._bookmarks, **kwparameters
+            self._config.impersonated_user, self._config.default_access_mode,
+            self._bookmarks, **kwparameters
         )
 
         return self._autoResult
@@ -261,21 +245,23 @@ class Session(Workspace):
             self._transaction = None
             self._disconnect()
 
-    def _transaction_network_error_handler(self):
+    def _transaction_error_handler(self, _):
         if self._transaction:
             self._transaction = None
             self._disconnect()
 
-    def _open_transaction(self, *, access_mode, database, metadata=None,
+    def _open_transaction(self, *, access_mode, metadata=None,
                           timeout=None):
-        self._connect(access_mode=access_mode, database=database)
+        self._connect(access_mode=access_mode)
         self._transaction = Transaction(
             self._connection, self._config.fetch_size,
             self._transaction_closed_handler,
-            self._transaction_network_error_handler
+            self._transaction_error_handler
+        )
+        self._transaction._begin(
+            self._config.database, self._config.impersonated_user,
+            self._bookmarks, access_mode, metadata, timeout
         )
-        self._transaction._begin(database, self._bookmarks, access_mode,
-                                 metadata, timeout)
 
     def begin_transaction(self, metadata=None, timeout=None):
         """ Begin a new unmanaged transaction. Creates a new :class:`.Transaction` within this session.
@@ -312,7 +298,8 @@ class Session(Workspace):
         if self._transaction:
             raise TransactionError("Explicit transaction already open")
 
-        self._open_transaction(access_mode=self._config.default_access_mode, database=self._config.database, metadata=metadata, timeout=timeout)
+        self._open_transaction(access_mode=self._config.default_access_mode,
+                               metadata=metadata, timeout=timeout)
 
         return self._transaction
 
@@ -332,7 +319,7 @@ class Session(Workspace):
 
         while True:
             try:
-                self._open_transaction(access_mode=access_mode, database=self._config.database, metadata=metadata, timeout=timeout)
+                self._open_transaction(access_mode=access_mode, metadata=metadata, timeout=timeout)
                 tx = self._transaction
                 try:
                     result = transaction_function(tx, *args, **kwargs)
