@@ -1,4 +1,4 @@
-# Copyright (c) 2002-2020 "Neo4j,"
+# Copyright (c) "Neo4j"
 # Neo4j Sweden AB [http://neo4j.com]
 #
 # This file is part of Neo4j.
@@ -15,43 +15,59 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-def toCypherAndParams(data):
-    cypher = data["cypher"]
+from neo4j.work.simple import Query
+
+
+def to_cypher_and_params(data):
+    from .backend import Request
     params = data["params"]
     # Optional
     if params is None:
-        return cypher, None
+        return data["cypher"], None
     # Transform the params to Python native
-    for p in params:
-        params[p] = toParam(params[p])
-    return cypher, params
+    params_dict = {p: to_param(params[p]) for p in params}
+    return data["cypher"], params_dict
 
 
-def toParam(m):
+def to_meta_and_timeout(data):
+    from .backend import Request
+    metadata = data.get('txMeta', None)
+    if isinstance(metadata, Request):
+        metadata.mark_all_as_read()
+    timeout = data.get('timeout', None)
+    if timeout:
+        timeout = timeout / 1000
+    return metadata, timeout
+
+
+def to_query_and_params(data):
+    cypher, param = to_cypher_and_params(data)
+    metadata, timeout = to_meta_and_timeout(data)
+    query = Query(cypher, metadata=metadata, timeout=timeout)
+    return query, param
+
+
+def to_param(m):
     """ Converts testkit parameter format to driver (python) parameter
     """
-    data = m["data"]
+    value = m["data"]["value"]
     name = m["name"]
-    if name == "CypherString":
-        return str(data["value"])
-    if name == "CypherBool":
-        return bool(data["value"])
-    if name == "CypherInt":
-        return int(data["value"])
-    if name == "CypherFloat":
-        return float(data["value"])
-    if name == "CypherString":
-        return str(data["value"])
     if name == "CypherNull":
         return None
+    if name == "CypherString":
+        return str(value)
+    if name == "CypherBool":
+        return bool(value)
+    if name == "CypherInt":
+        return int(value)
+    if name == "CypherFloat":
+        return float(value)
+    if name == "CypherString":
+        return str(value)
+    if name == "CypherBytes":
+        return bytearray([int(byte, 16) for byte in value.split()])
     if name == "CypherList":
-        ls = []
-        for x in data["value"]:
-            ls.append(toParam(x))
-        return ls
+        return [to_param(v) for v in value]
     if name == "CypherMap":
-        mp = {}
-        for k, v in data["value"].items():
-            mp[k] = toParam(v)
-        return mp
+        return {k: to_param(value[k]) for k in value}
     raise Exception("Unknown param type " + name)
